#include <libsmbclient.h>
#include <pthread.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include <popt.h>

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

static void auth_smbc_get_data( __attribute__ ((unused)) const char *server,
			       __attribute__ ((unused)) const char *share,
			       __attribute__ ((unused)) char *workgroup,
			       __attribute__ ((unused)) int wgmaxlen,
			       char *username, __attribute__ ((unused)) int unmaxlen,
			       char *password, __attribute__ ((unused)) int pwmaxlen)
{
    char cred_file [1024];
    FILE *fp;

    snprintf(cred_file, 1024, "%s/.smb/credentials", getenv("HOME"));
    fp = fopen(cred_file, "r");
    if( fp != NULL )
    {
        char fbuf[1024];
        while( fgets( fbuf, 1024, fp ) != NULL )
        {
            if( strncmp( fbuf, "username", 8 ) == 0 )
            {
                char* begin = fbuf + 8;
                char* end;
                while ( *begin == ' ' || *begin == '\t')
                     begin++;
                if (*begin == '=')
                {
                    begin++;
                    while ( *begin == ' ' || *begin == '\t')
                        begin++;
                    strncpy( username, begin, strlen(begin));
                    end = username + strlen(username) - 1;
                    while ( *end == ' ' || *end == '\t' || *end == '\n' || *end == '\0')
                    {
                        end--;
                    }
                    end++;
                    *end = '\0';

                }
            }
            else if( strncmp( fbuf, "password", 8 ) == 0 )
            {
                char* begin = fbuf + 8;
                char* end;
                while ( *begin == ' ' || *begin == '\t')
                     begin++;
                if (*begin == '=')
                {
                    begin++;
                    while ( *begin == ' ' || *begin == '\t')
                        begin++;
                    strncpy( password, begin, strlen(begin));
                    end = password + strlen(password) - 1;
                    while ( *end == ' ' || *end == '\t' || *end == '\n' || *end == '\0')
                    {
                        end--;
                    }
                    end++;
                    *end = '\0';
                }
            }
        }
        if (!strlen(username))
        {
            char un[] = "guest";
            char pw[] = "";

            strncpy(username, un, sizeof(un));
            strncpy(password, pw, sizeof(pw));
        }
    }
    else
    {
        char un[] = "guest";
        char pw[] = "";

        strncpy(username, un, sizeof(un));
        strncpy(password, pw, sizeof(pw));
    }
}

typedef struct smbcrawler_options {
    SMBCCTX *smb_ctx;
    FILE *file;
    const char *separator;
    int use_file;
    int show_comments;
} smbcrawler_options_t;

void smbcrawler_init (smbcrawler_options_t *options, const char *output_file, const char *separator, int timeout, int debuglevel)
{
    options->smb_ctx = smbc_new_context();
    options->smb_ctx->callbacks.auth_fn = auth_smbc_get_data;
    /* Timeout a bit bigger, by Jim Ramsay */
    options->smb_ctx->timeout = timeout * 1000;
    options->smb_ctx->debug = debuglevel;
    /* Kerberos authentication by Esben Nielsen */
#if defined(SMB_CTX_FLAG_USE_KERBEROS) && defined(SMB_CTX_FLAG_FALLBACK_AFTER_KERBEROS)
    options->smb_ctx->flags |= SMB_CTX_FLAG_USE_KERBEROS|SMB_CTX_FLAG_FALLBACK_AFTER_KERBEROS;
#endif
    options->smb_ctx = smbc_init_context(options->smb_ctx);
    if (output_file == NULL)
    {
        options->use_file = 0;
        options->file = fopen(output_file, "w");
        if (!options->file)
        {
            fprintf(stderr, "Could not open file: %s\n", output_file);
            smbc_free_context(options->smb_ctx, 1);
            exit(1);
        }
    }
    else
    {
        options->use_file = 1;
        options->file = stdout;
    }
    options->separator = separator;
}

void smbcrawler_close(smbcrawler_options_t *options)
{
    if (options->use_file)
    {
        fclose(options->file);
    }
    smbc_free_context(options->smb_ctx, 1);
}

void smbcrawler_share(smbcrawler_options_t *options, const char *server, const char *share, const char *path, int level)
{
    SMBCFILE *dir;
    struct stat st;
    struct smbc_dirent *share_dirent;
    char path_name[8192];
    strcpy(path_name, "smb://");
    strcat(path_name, server);
    strcat(path_name, "/");
    strcat(path_name, share);
    if (strlen(path))
    {
        strcat(path_name, "/");
        strcat(path_name, path);
    }
    dir = options->smb_ctx->opendir(options->smb_ctx, path_name);

    //printf("%s %s %s %s\n", server, share, path, path_name);
    if (dir)
    {
        while ((share_dirent = options->smb_ctx->readdir(options->smb_ctx, dir)) != NULL)
        {
            if (share_dirent->smbc_type == SMBC_FILE)
            {
                char stat_file_name[8192];
                strcpy(stat_file_name, path_name);
                strcat(stat_file_name, "/");
                strcat(stat_file_name, share_dirent->name);
                options->smb_ctx->stat(options->smb_ctx, stat_file_name, &st);

                fprintf(options->file, "%s%s%s%s%s%s%Lu\n", share,
                                                         options->separator,
                                                         path,
                                                         options->separator,
                                                         share_dirent->name,
                                                         options->separator,
                                                         st.st_size);
            }
            else
            {
                //printf("DIR: %s\n", share_dirent->name);
                if (!(strcmp(share_dirent->name, ".") == 0 ||
                    strcmp(share_dirent->name, "..") == 0))
                {
                    //printf("DIR: %s\n", share_dirent->name);
                    char buff[8192];
                    strcpy(buff, path);
                    if (strlen(path))
                    {
                        strcpy(buff, path);
                        strcat(buff, "/");
                    }
                    strcat(buff, share_dirent->name);
                    smbcrawler_share(options, server, share, buff, 0);
                }
            }
        }
        options->smb_ctx->closedir(options->smb_ctx, dir);
    }

}


void smbcrawler_server(smbcrawler_options_t *options, const char *server)
{
    
    char computer[4096] = "smb://";
    SMBCFILE *dir;
    struct smbc_dirent *server_dirent;
    
    strcat(computer, server);
    dir = options->smb_ctx->opendir(options->smb_ctx, computer);
    if (dir)
    {
        while ((server_dirent = options->smb_ctx->readdir(options->smb_ctx, dir)) != NULL)
        {
            if (server_dirent->namelen > 0 ||
                server_dirent->smbc_type == SMBC_FILE_SHARE)
            {
                if (server_dirent->name[strlen(server_dirent->name) -1] != '$')
                {
                    char buff[4096];
                    strcpy(buff, server_dirent->name);
                    smbcrawler_share(options, server, buff, "", 0);
                }
            }
        }
        options->smb_ctx->closedir(options->smb_ctx, dir);
    }
    else
    {
        fprintf(stderr, "Could not find server: %s\n", server);
    }
}

void smbcrawler_servers (smbcrawler_options_t *options)
{
    char wg[] = "smb://";
    SMBCFILE *dir, *server_dir;
    struct smbc_dirent *wg_dirent, *server_dirent;
    
    dir = options->smb_ctx->opendir(options->smb_ctx, wg);
    if (dir)
    {
        while ((wg_dirent = options->smb_ctx->readdir(options->smb_ctx, dir)) != NULL)
        {
            if (wg_dirent->namelen > 0 ||
                wg_dirent->smbc_type == SMBC_WORKGROUP)
            {
                char buff[4096] = "smb://";
                strcat(buff, wg_dirent->name);
                server_dir = options->smb_ctx->opendir(options->smb_ctx, buff);
                if (server_dir)
                {
                    while ((server_dirent = options->smb_ctx->readdir(options->smb_ctx, server_dir)) != NULL)
                    {
                        if (server_dirent->namelen > 0 ||
                            server_dirent->smbc_type == SMBC_SERVER)
                        {
                            if (!options->show_comments)
                            {
                                fprintf(options->file, "%s%s%s\n", buff+6,
                                                              options->separator,
                                                              server_dirent->name);
                            }
                            else
                            {
                                fprintf(options->file, "%s%s%s%s%s\n", buff+6,
                                        options->separator,
                                        server_dirent->name,
                                        options->separator,
                                        server_dirent->comment
                                       );
                            }
                        }
                    }
                    options->smb_ctx->closedir(options->smb_ctx, server_dir);
                }
            }
        }
        options->smb_ctx->closedir(options->smb_ctx, dir);
    }
}

void smbcrawler_shares (smbcrawler_options_t *options, const char *server)
{
    SMBCFILE *server_dir;
    struct smbc_dirent *server_dirent;
    
                char buff[4096] = "smb://";
                strcat(buff, server);
                server_dir = options->smb_ctx->opendir(options->smb_ctx, buff);
                if (server_dir)
                {
                    while ((server_dirent = options->smb_ctx->readdir(options->smb_ctx, server_dir)) != NULL)
                    {
                        if (server_dirent->namelen > 0 ||
                            server_dirent->smbc_type == SMBC_FILE_SHARE)
                        {
                            if (options->show_comments)
                                fprintf(options->file, "%s%s%s\n", server_dirent->name, options->separator, server_dirent->comment);
                            else
                                fprintf(options->file, "%s\n", server_dirent->name);
                        }
                    }
                    options->smb_ctx->closedir(options->smb_ctx, server_dir);
                }
}


void smbcrawler_workgroups (smbcrawler_options_t *options)
{
    char wg[] = "smb://";
    SMBCFILE *dir;
    struct smbc_dirent *wg_dirent;
    
    dir = options->smb_ctx->opendir(options->smb_ctx, wg);
    if (dir)
    {
        while ((wg_dirent = options->smb_ctx->readdir(options->smb_ctx, dir)) != NULL)
        {
            if (wg_dirent->namelen > 0 ||
                wg_dirent->smbc_type == SMBC_WORKGROUP)
            {
                fprintf(options->file, "%s\n", wg_dirent->name);
            }
        }
        options->smb_ctx->closedir(options->smb_ctx, dir);
    }
}


int main(int argc, const char *argv[])
{
    const char *output;
    const char *server_name;
    const char *server;
    const char *separator = ":";
    smbcrawler_options_t options;
    
    int crawl_workgroups=0,
        crawl_servers = 0,
        crawl_server = 0,
        crawl_shares =0,
        use_file =0,
        show_comments = 0;
    char c;
    
    poptContext opt_ctx;
    struct poptOption opt_table[] = {
        {"workgroups", 'w', 0, 0, 'w',
        "List available workgroups"},
        {"servers", 's', 0, 0, 's',
         "List available servers"},
        {"shares", 'l', POPT_ARG_STRING, &server, 'l',
         "List available shares", "SERVER"},
        {"comments", 'c', 0, 0, 'c',
         "Show comments when listing shares or servers"},
        {"output", 'o', POPT_ARG_STRING, &output, 'o',
         "Output file name (default=stdout)", "FILE"},
        {"field-separator", 'f', POPT_ARG_STRING, &separator, 'f',
         "Field separator (default= : )", "SEP"},
        {"help", 'h', 0, 0, 'h',
         "Show this help message"},
        {"usage", '\0', 0, 0, 'u',
         "Display brief usage message"},
        {"version", 'V', 0, 0, 'V',
         "Print version"},
         { NULL, '\0', 0, NULL, 0, NULL, NULL }
    };
    opt_ctx = poptGetContext(NULL, argc, argv, opt_table, 0);
    poptSetOtherOptionHelp(opt_ctx, "[options]* [server_name]");
    
    if (argc < 2)
    {
        poptPrintHelp(opt_ctx, stderr, 0);
        exit(1);
    }
    while ((c = poptGetNextOpt(opt_ctx)) >= 0)
    {
        switch (c) {
            case 'w':
                crawl_workgroups = 1;
                break;
            case 's':
                crawl_servers = 1;
                break;
            case 'o':
                use_file = 1;
                break;
            case 'l':
                crawl_shares = 1;
                break;
            case 'c':
                show_comments = 1;
                break;
            case 'V':
                printf("Version %s\n", VERSION);
                exit(0);
                break;
            case 'h':
                poptPrintHelp(opt_ctx, stderr, 0);
                exit(1);
                break;
            case 'u':
                poptPrintUsage(opt_ctx, stderr, 0);
                exit(1);
                break;
        }
    }
    server_name = poptGetArg(opt_ctx);
    if (server_name != NULL)
    {
        if (crawl_workgroups)
        {
            fprintf(stderr, "Cannot list workgroups when <server_name> is specified\n");
            exit(1);
        }
        if (crawl_servers)
        {
            fprintf(stderr, "Cannot list servers when <server_name> is specified\n");
            exit(1);
        }
        if (crawl_shares)
        {
            fprintf(stderr, "Cannot list shares when <server_name> is specified\n");
            exit(1);
        }
        crawl_server = 1;
    }
    else if (crawl_servers)
    {
        if (crawl_workgroups)
        {
            fprintf(stderr, "Cannot scan workgroups and servers at the same time\n");
            exit(1);
        }
    }
    else if (crawl_workgroups)
    {

    }
    else if (crawl_shares)
    {
        
    }
    else if (server_name == NULL ||
        !(poptPeekArg(opt_ctx) == NULL))
    {
        poptPrintHelp(opt_ctx, stderr, 0);
        fprintf(stderr, "Specify a (single) <server_name>\n");
        exit(1);
    }
    options.show_comments = show_comments;
    smbcrawler_init(&options, output, separator, 10, 0);

    if (crawl_servers)
    {
        smbcrawler_servers(&options);
    }
    if (crawl_workgroups)
    {
        smbcrawler_workgroups(&options);
    }
    if (crawl_server)
    {
        smbcrawler_server (&options, server_name);
    }
    if (crawl_shares)
    {
        smbcrawler_shares (&options, server);
    }
    smbcrawler_close(&options);
    exit(0);
}
